package com.sun.xml.internal.ws.server.provider;

import com.sun.xml.internal.ws.server.AbstractWebServiceContext;
import com.sun.istack.internal.Nullable;
import com.sun.xml.internal.ws.api.server.AsyncProviderCallback;
import com.sun.xml.internal.ws.api.pipe.Fiber;
import com.sun.xml.internal.ws.api.server.WSEndpoint;
import com.sun.xml.internal.ws.api.pipe.ThrowableContainerPropertySet;
import java.util.logging.Level;
import com.sun.xml.internal.ws.api.pipe.NextAction;
import com.sun.istack.internal.NotNull;
import com.sun.xml.internal.ws.api.message.Packet;
import com.sun.xml.internal.ws.api.server.Invoker;
import java.util.logging.Logger;

public class AsyncProviderInvokerTube<T> extends ProviderInvokerTube<T>
{
    private static final Logger LOGGER;
    
    public AsyncProviderInvokerTube(Invoker invoker, ProviderArgumentsBuilder<T> argsBuilder) {
        super(invoker, argsBuilder);
    }
    
    @NotNull
    public NextAction processRequest(@NotNull Packet request) {
        T param = this.argsBuilder.getParameter(request);
        AsyncProviderInvokerTube.NoSuspendResumer resumer = new AsyncProviderInvokerTube.NoSuspendResumer();
        AsyncProviderInvokerTube.AsyncProviderCallbackImpl callback = new AsyncProviderInvokerTube.AsyncProviderCallbackImpl(request, resumer);
        AsyncProviderInvokerTube.AsyncWebServiceContext ctxt = new AsyncProviderInvokerTube.AsyncWebServiceContext(this.getEndpoint(), request);
        AsyncProviderInvokerTube.LOGGER.fine("Invoking AsyncProvider Endpoint");
        try {
            this.getInvoker(request).invokeAsyncProvider(request, param, callback, ctxt);
        }
        catch (Throwable e) {
            AsyncProviderInvokerTube.LOGGER.log(Level.SEVERE, e.getMessage(), e);
            return this.doThrow(e);
        }
        synchronized (callback) {
            if (resumer.response != null) {
                ThrowableContainerPropertySet tc = (ThrowableContainerPropertySet)resumer.response.getSatellite(ThrowableContainerPropertySet.class);
                Throwable t = (tc != null) ? tc.getThrowable() : null;
                return (t != null) ? this.doThrow(resumer.response, t) : this.doReturnWith(resumer.response);
            }
            callback.resumer = new AsyncProviderInvokerTube.FiberResumer();
            return this.doSuspend();
        }
    }
    
    @NotNull
    public NextAction processResponse(@NotNull Packet response) {
        return this.doReturnWith(response);
    }
    
    @NotNull
    public NextAction processException(@NotNull Throwable t) {
        return this.doThrow(t);
    }
    
    static WSEndpoint access$200(AsyncProviderInvokerTube x0) {
        return x0.getEndpoint();
    }
    
    static WSEndpoint access$300(AsyncProviderInvokerTube x0) {
        return x0.getEndpoint();
    }
    
    static WSEndpoint access$400(AsyncProviderInvokerTube x0) {
        return x0.getEndpoint();
    }
    
    static WSEndpoint access$500(AsyncProviderInvokerTube x0) {
        return x0.getEndpoint();
    }
    
    static {
        LOGGER = Logger.getLogger("com.sun.xml.internal.ws.server.AsyncProviderInvokerTube");
    }
    
    public class FiberResumer implements AsyncProviderInvokerTube.Resumer
    {
        private final Fiber fiber;
        
        public FiberResumer() {
            this.fiber = Fiber.current();
        }
        
        @Override
        public void onResume(Packet response) {
            ThrowableContainerPropertySet tc = (ThrowableContainerPropertySet)response.getSatellite(ThrowableContainerPropertySet.class);
            Throwable t = (tc != null) ? tc.getThrowable() : null;
            this.fiber.resume(t, response);
        }
    }
    
    private class NoSuspendResumer implements AsyncProviderInvokerTube.Resumer
    {
        protected Packet response;
        
        private NoSuspendResumer() {
            this.response = null;
        }
        
        @Override
        public void onResume(Packet response) {
            this.response = response;
        }
    }
    
    public class AsyncProviderCallbackImpl implements AsyncProviderCallback<T>
    {
        private final Packet request;
        private AsyncProviderInvokerTube.Resumer resumer;
        
        public AsyncProviderCallbackImpl(Packet request, AsyncProviderInvokerTube.Resumer resumer) {
            this.request = request;
            this.resumer = resumer;
        }
        
        public void send(@Nullable T param) {
            if (param == null && this.request.transportBackChannel != null) {
                this.request.transportBackChannel.close();
            }
            Packet packet = AsyncProviderInvokerTube.this.argsBuilder.getResponse(this.request, param, AsyncProviderInvokerTube.access$200(AsyncProviderInvokerTube.this).getPort(), AsyncProviderInvokerTube.access$300(AsyncProviderInvokerTube.this).getBinding());
            synchronized (this) {
                this.resumer.onResume(packet);
            }
        }
        
        public void sendError(@NotNull Throwable t) {
            Exception e;
            if ((t instanceof Exception)) {
                e = (Exception)t;
            }
            else {
                e = new RuntimeException(t);
            }
            Packet packet = AsyncProviderInvokerTube.this.argsBuilder.getResponse(this.request, e, AsyncProviderInvokerTube.access$400(AsyncProviderInvokerTube.this).getPort(), AsyncProviderInvokerTube.access$500(AsyncProviderInvokerTube.this).getBinding());
            synchronized (this) {
                this.resumer.onResume(packet);
            }
        }
    }
    
    public class AsyncWebServiceContext extends AbstractWebServiceContext
    {
        final Packet packet;
        
        public AsyncWebServiceContext(WSEndpoint endpoint, Packet packet) {
            super(endpoint);
            this.packet = packet;
        }
        
        @NotNull
        public Packet getRequestPacket() {
            return this.packet;
        }
    }
    
    private interface Resumer
    {
        public abstract void onResume(Packet p0);
    }
}
